import json

from django.contrib.auth.decorators import login_required, permission_required
from django.http import HttpResponse, JsonResponse
from django.shortcuts import render
from django.utils.html import format_html

from app_utils.views import link_html, no_wrap_html, yesno_str

from . import __title__
from .app_settings import (
    PACKAGE_MONITOR_INCLUDE_PACKAGES,
    PACKAGE_MONITOR_SHOW_ALL_PACKAGES,
)
from .models import Distribution

PACKAGE_LIST_FILTER_PARAM = "filter"


@login_required
@permission_required("package_monitor.basic_access")
def index(request):
    obj = Distribution.objects.first()
    updated_at = obj.updated_at if obj else None
    distributions_qs = Distribution.objects.currently_selected()
    filter = request.GET.get(PACKAGE_LIST_FILTER_PARAM)
    if not filter:
        app_count = Distribution.objects.currently_selected().outdated_count()
        filter = "outdated" if app_count and app_count > 0 else "current"
    outdated_install_command = (
        Distribution.objects.currently_selected()
        .filter(is_outdated=True)
        .order_by("name")
        .build_install_command()
    )
    context = {
        "app_title": __title__,
        "page_title": "Distribution packages",
        "updated_at": updated_at,
        "filter": filter,
        "all_count": distributions_qs.count(),
        "current_count": distributions_qs.filter(is_outdated=False).count(),
        "outdated_count": distributions_qs.outdated_count(),
        "unknown_count": distributions_qs.filter(is_outdated__isnull=True).count(),
        "include_packages": PACKAGE_MONITOR_INCLUDE_PACKAGES,
        "show_all_packages": PACKAGE_MONITOR_SHOW_ALL_PACKAGES,
        "outdated_install_command": outdated_install_command,
    }
    return render(request, "package_monitor/index.html", context)


@login_required
@permission_required("package_monitor.basic_access")
def package_list_data(request) -> JsonResponse:
    """Returns the packages as list in JSON.
    Specify different subsets with the "filter" GET parameter
    """
    my_filter = request.GET.get(PACKAGE_LIST_FILTER_PARAM, "")
    distributions_qs = Distribution.objects.currently_selected()

    if my_filter == "outdated":
        distributions_qs = distributions_qs.filter(is_outdated=True)
    elif my_filter == "current":
        distributions_qs = distributions_qs.filter(is_outdated=False)
    elif my_filter == "unknown":
        distributions_qs = distributions_qs.filter(is_outdated__isnull=True)

    data = list()
    for dist in distributions_qs.order_by("name"):
        name_link_html = (
            link_html(dist.website_url, dist.name) if dist.website_url else dist.name
        )
        if dist.is_outdated:
            name_link_html += '&nbsp;<i class="fas fa-exclamation-circle" title="Update available"></i>'

        if dist.apps:
            _lst = [no_wrap_html(row) for row in json.loads(dist.apps)]
            apps_html = "<br>".join(_lst) if _lst else "-"
        else:
            apps_html = ""

        if dist.used_by:
            used_by_sorted = sorted(json.loads(dist.used_by), key=lambda k: k["name"])
            used_by_html = "<br>".join(
                [
                    format_html(
                        '<span title="{}" class="text-nowrap;">{}</span>',
                        ", ".join(row["requirements"])
                        if row["requirements"]
                        else "ANY",
                        link_html(row["homepage_url"], row["name"])
                        if row["homepage_url"]
                        else row["name"],
                    )
                    for row in used_by_sorted
                ]
            )
        else:
            used_by_html = ""

        if not dist.latest_version:
            latest_html = "?"
        else:
            command = f"pip install {dist.pip_install_version}"
            latest_html = no_wrap_html(
                f'<span class="copy_to_clipboard" '
                f'title="{command}"'
                f' data-clipboard-text="{command}">'
                f"{dist.latest_version}"
                '&nbsp;&nbsp;<i class="far fa-copy"></i></span>'
            )

        data.append(
            {
                "name": dist.name,
                "name_link": no_wrap_html(name_link_html),
                "apps": apps_html,
                "used_by": used_by_html,
                "current": dist.installed_version,
                "latest": latest_html,
                "is_outdated": dist.is_outdated,
                "is_outdated_str": yesno_str(dist.is_outdated),
                "description": dist.description,
            }
        )

    return JsonResponse(data, safe=False)


@login_required
@permission_required("package_monitor.basic_access")
def refresh_distributions(request):
    Distribution.objects.update_all(use_threads=True)
    return HttpResponse("ok")



from django.urls import path

from . import views

app_name = "package_monitor"

urlpatterns = [
    path("", views.index, name="index"),
    path("package_list_data", views.package_list_data, name="package_list_data"),
    path(
        "refresh_distributions",
        views.refresh_distributions,
        name="refresh_distributions",
    ),
]



import json
from unittest.mock import patch

from django.contrib.auth.models import User
from django.test import RequestFactory, TestCase
from django.urls import reverse

from allianceauth.tests.auth_utils import AuthUtils

from .. import views
from .testdata import create_testdata

MODULE_PATH_VIEWS = "package_monitor.views"
MODULE_PATH_MANAGERS = "package_monitor.managers"


@patch(MODULE_PATH_MANAGERS + ".PACKAGE_MONITOR_SHOW_ALL_PACKAGES", True)
@patch(MODULE_PATH_MANAGERS + ".PACKAGE_MONITOR_INCLUDE_PACKAGES", None)
class TestPackageList(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        create_testdata()
        cls.user = AuthUtils.create_user("Bruce Wayne")
        AuthUtils.add_permission_to_user_by_name(
            "package_monitor.basic_access", cls.user
        )
        cls.user = User.objects.get(pk=cls.user.pk)

    def setUp(self) -> None:
        self.factory = RequestFactory()

    def test_index_view(self):
        request = self.factory.get(reverse("package_monitor:index"))
        request.user = self.user
        response = views.index(request)
        self.assertEqual(response.status_code, 200)

    def test_list_view_all(self):
        request = self.factory.get(reverse("package_monitor:package_list_data"))
        request.user = self.user
        response = views.package_list_data(request)
        self.assertEqual(response.status_code, 200)

        data = json.loads(response.content.decode("utf-8"))
        package_names = [x["name"] for x in data]
        self.assertListEqual(package_names, ["dummy-1", "dummy-2", "dummy-3"])

    def test_list_view_outdated(self):
        request = self.factory.get(
            reverse("package_monitor:package_list_data"), data={"filter": "outdated"}
        )
        request.user = self.user
        response = views.package_list_data(request)
        self.assertEqual(response.status_code, 200)

        data = json.loads(response.content.decode("utf-8"))
        package_names = [x["name"] for x in data]
        self.assertListEqual(package_names, ["dummy-2"])

    def test_list_view_current(self):
        request = self.factory.get(
            reverse("package_monitor:package_list_data"), data={"filter": "current"}
        )
        request.user = self.user
        response = views.package_list_data(request)
        self.assertEqual(response.status_code, 200)

        data = json.loads(response.content.decode("utf-8"))
        package_names = [x["name"] for x in data]
        self.assertListEqual(package_names, ["dummy-1"])

    def test_list_view_unknown(self):
        request = self.factory.get(
            reverse("package_monitor:package_list_data"), data={"filter": "unknown"}
        )
        request.user = self.user
        response = views.package_list_data(request)
        self.assertEqual(response.status_code, 200)

        data = json.loads(response.content.decode("utf-8"))
        package_names = [x["name"] for x in data]
        self.assertListEqual(package_names, ["dummy-3"])

    @patch(MODULE_PATH_VIEWS + ".Distribution.objects.update_all")
    def test_refresh_distributions_view(self, mock_update_all):
        mock_update_all.return_value = 1

        request = self.factory.get(reverse("package_monitor:refresh_distributions"))
        request.user = self.user
        response = views.refresh_distributions(request)
        self.assertEqual(response.status_code, 200)



# Generated by Django 2.2.16 on 2020-09-16 18:39

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (("basic_access", "Can access this app"),),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Distribution",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Name of this package", max_length=255, unique=True
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        default="", help_text="Description of this package"
                    ),
                ),
                (
                    "apps",
                    models.TextField(
                        default="",
                        help_text="List of installed Django apps included in this package (JSON)",
                    ),
                ),
                (
                    "has_installed_apps",
                    models.BooleanField(
                        db_index=True,
                        default=False,
                        help_text="Whether this package has installed Django apps",
                    ),
                ),
                (
                    "used_by",
                    models.TextField(
                        default="",
                        help_text="List of other distribution packages using this package (JSON)",
                    ),
                ),
                (
                    "installed_version",
                    models.CharField(
                        default="",
                        help_text="Currently installed version of this package",
                        max_length=64,
                    ),
                ),
                (
                    "latest_version",
                    models.CharField(
                        default="",
                        help_text="Latest stable version available for this package",
                        max_length=64,
                    ),
                ),
                (
                    "is_outdated",
                    models.BooleanField(
                        db_index=True,
                        default=None,
                        help_text="A package is outdated when there is a newer stable version available",
                        null=True,
                    ),
                ),
                (
                    "website_url",
                    models.TextField(
                        default="", help_text="URL to the home page of this package"
                    ),
                ),
                (
                    "updated_at",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="Date & time this data was last updated",
                    ),
                ),
            ],
        ),
    ]



from celery import shared_task

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .models import Distribution

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@shared_task(time_limit=3600)
def update_distributions():
    Distribution.objects.update_all()



import json

import factory
import factory.fuzzy

from package_monitor.models import Distribution


class DistributionFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Distribution
        django_get_or_create = ("name",)

    name = factory.Faker("last_name")
    description = factory.Faker("paragraph")
    latest_version = factory.LazyAttribute(lambda o: o.installed_version)
    is_outdated = False
    website_url = factory.Faker("uri")

    class Params:
        app_list = []

    @factory.lazy_attribute
    def installed_version(self):
        int_fuzzer = factory.fuzzy.FuzzyInteger(0, 20)
        major = int_fuzzer.fuzz()
        minor = int_fuzzer.fuzz()
        patch = int_fuzzer.fuzz()
        return f"{major}.{minor}.{patch}"

    @factory.lazy_attribute
    def apps(self):
        return json.dumps(self.app_list)



from .factories import DistributionFactory


def create_testdata():
    DistributionFactory(
        name="dummy-2",
        installed_version="0.1.0",
        latest_version="0.2.0",
        is_outdated=True,
    )
    DistributionFactory(
        name="dummy-1",
        app_list=["app_1"],
        installed_version="0.1.0",
        latest_version="0.1.0",
        is_outdated=False,
    )
    DistributionFactory(name="dummy-3", installed_version="0.1.0", is_outdated=None)



import json
from typing import Dict

import importlib_metadata
from packaging.utils import canonicalize_name
from packaging.version import parse as version_parse

from django.db import models, transaction

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import (
    PACKAGE_MONITOR_INCLUDE_PACKAGES,
    PACKAGE_MONITOR_SHOW_ALL_PACKAGES,
)
from .core import (
    DistributionPackage,
    compile_package_requirements,
    fetch_relevant_packages,
    update_packages_from_pypi,
)

TERMINAL_MAX_LINE_LENGTH = 4095

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def _none_2_empty(text) -> str:
    """Translate None to empty string."""
    if text is None:
        return ""
    return text


class DistributionQuerySet(models.QuerySet):
    def outdated_count(self) -> int:
        return self.filter(is_outdated=True).count()

    def build_install_command(self) -> str:
        """Build install command from all distribution packages in this query."""
        result = "pip install"
        for dist in self.exclude(latest_version=""):
            version_string = dist.pip_install_version
            if len(result) + len(version_string) + 1 > TERMINAL_MAX_LINE_LENGTH:
                break
            result = f"{result} {version_string}"
        return result


class DistributionManagerBase(models.Manager):
    def currently_selected(self) -> models.QuerySet:
        """Currently selected packages based on global settings,
        e.g. related to installed apps vs. all packages
        """
        if PACKAGE_MONITOR_SHOW_ALL_PACKAGES:
            return self.all()
        qs = self.filter(has_installed_apps=True)
        if PACKAGE_MONITOR_INCLUDE_PACKAGES:
            qs |= self.filter(name__in=PACKAGE_MONITOR_INCLUDE_PACKAGES)
        return qs

    def update_all(self, use_threads=False) -> int:
        """Updates the list of relevant distribution packages in the database"""
        logger.info(
            f"Started refreshing approx. {self.count()} distribution packages..."
        )
        packages = fetch_relevant_packages(importlib_metadata.distributions())
        requirements = compile_package_requirements(
            packages, importlib_metadata.distributions()
        )
        update_packages_from_pypi(packages, requirements, use_threads)
        self._save_packages(packages, requirements)
        packages_count = len(packages)
        logger.info(f"Completed refreshing {packages_count} distribution packages")
        return packages_count

    def _save_packages(
        self, packages: Dict[str, DistributionPackage], requirements: dict
    ) -> None:
        """Saves the given package information into the model"""

        def metadata_value(dist, prop: str) -> str:
            return (
                dist.metadata[prop]
                if dist and dist.metadata.get(prop) != "UNKNOWN"
                else ""
            )

        def packages_lookup(packages: dict, name: str, attr: str, default=None):
            package = packages.get(canonicalize_name(name))
            return getattr(package, attr) if package else default

        with transaction.atomic():
            self.all().delete()
            objs = list()
            for package_name, package in packages.items():
                is_outdated = (
                    version_parse(package.current) < version_parse(package.latest)
                    if package.current
                    and package.latest
                    and str(package.current) == str(package.distribution.version)
                    else None
                )
                if package_name in requirements:
                    used_by = [
                        {
                            "name": package_name,
                            "homepage_url": metadata_value(
                                packages_lookup(packages, package_name, "distribution"),
                                "Home-page",
                            ),
                            "requirements": [str(obj) for obj in package_requirements],
                        }
                        for package_name, package_requirements in requirements[
                            package_name
                        ].items()
                    ]
                else:
                    used_by = []

                name = _none_2_empty(package.distribution.metadata["Name"])
                apps = _none_2_empty(json.dumps(sorted(package.apps, key=str.casefold)))
                used_by = _none_2_empty(json.dumps(used_by))
                installed_version = _none_2_empty(package.distribution.version)
                latest_version = str(package.latest) if package.latest else ""
                description = _none_2_empty(
                    metadata_value(package.distribution, "Summary")
                )
                website_url = _none_2_empty(
                    metadata_value(package.distribution, "Home-page")
                )
                obj = self.model(
                    name=name,
                    apps=apps,
                    used_by=used_by,
                    installed_version=installed_version,
                    latest_version=latest_version,
                    is_outdated=is_outdated,
                    description=description,
                    website_url=website_url,
                )
                obj.calc_has_installed_apps()
                objs.append(obj)
            self.bulk_create(objs)


DistributionManager = DistributionManagerBase.from_queryset(DistributionQuerySet)



import concurrent.futures
import sys
from dataclasses import dataclass, field
from typing import Dict, Iterable, List

import importlib_metadata
import requests
from packaging.markers import UndefinedComparison, UndefinedEnvironmentName
from packaging.requirements import InvalidRequirement, Requirement
from packaging.specifiers import SpecifierSet
from packaging.utils import canonicalize_name
from packaging.version import parse as version_parse

from django.apps import apps as django_apps

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

# max workers used when fetching info from PyPI for packages
MAX_THREAD_WORKERS = 30


@dataclass
class DistributionWrapped:
    """Distribution with some additional information."""

    name: str
    distribution: importlib_metadata.Distribution
    files: List[str] = field(default_factory=list)

    @classmethod
    def from_distribution(
        cls, dist: importlib_metadata.Distribution
    ) -> "DistributionWrapped":
        return cls(
            name=canonicalize_name(dist.metadata["Name"]),
            distribution=dist,
            files=["/" + str(f) for f in dist.files if str(f).endswith("__init__.py")],
        )

    @classmethod
    def from_distributions(
        cls, distributions: Iterable[importlib_metadata.Distribution]
    ) -> "DistributionWrapped":
        return [
            DistributionWrapped.from_distribution(dist)
            for dist in distributions
            if dist.metadata["Name"]
        ]


@dataclass
class DistributionPackage:
    """A distribution package."""

    name: str
    current: str
    distribution: importlib_metadata.Distribution
    requirements: List[Requirement] = field(default_factory=list)
    apps: List[str] = field(default_factory=list)
    latest: str = ""


def fetch_relevant_packages(
    distributions: Iterable[importlib_metadata.Distribution],
) -> Dict[str, DistributionPackage]:
    """Fetch distribution packages with packages relevant for this Django installation"""
    packages = dict()
    for dist in DistributionWrapped.from_distributions(distributions):
        if dist.name not in packages:
            packages[dist.name] = DistributionPackage(
                **{
                    "name": dist.name,
                    "current": dist.distribution.version,
                    "requirements": _parse_requirements(dist.distribution.requires),
                    "distribution": dist.distribution,
                }
            )
        for dist_file in dist.files:
            for app in django_apps.get_app_configs():
                my_file = app.module.__file__
                if my_file.endswith(dist_file):
                    packages[dist.name].apps.append(app.name)
                    break
    return packages


def _parse_requirements(requires: list) -> List[Requirement]:
    """Parse requirements from a distribution and return them.

    Invalid requirements will be ignored.
    """
    requirements = list()
    if requires:
        for r in requires:
            try:
                requirements.append(Requirement(r))
            except InvalidRequirement:
                pass
    return requirements


def compile_package_requirements(
    packages: Dict[str, DistributionPackage],
    distributions: Iterable[importlib_metadata.Distribution],
) -> dict:
    """Consolidate requirements from all known distributions and known packages"""
    requirements = dict()
    for dist in distributions:
        if dist.requires:
            for requirement in _parse_requirements(dist.requires):
                requirement_name = canonicalize_name(requirement.name)
                if requirement_name in packages:
                    if requirement.marker:
                        try:
                            is_valid = requirement.marker.evaluate()
                        except (UndefinedEnvironmentName, UndefinedComparison):
                            is_valid = False
                    else:
                        is_valid = True

                    if is_valid:
                        if requirement_name not in requirements:
                            requirements[requirement_name] = dict()

                        requirements[requirement_name][
                            dist.metadata["Name"]
                        ] = requirement.specifier

    return requirements


def update_packages_from_pypi(
    packages: Dict[str, DistributionPackage], requirements: dict, use_threads=False
) -> None:
    """Fetch the latest versions for given packages from PyPI in accordance
    with the given requirements and updates the packages.
    """

    def thread_update_latest_from_pypi(package_name: str) -> None:
        """Retrieves latest valid version from PyPI and updates packages

        Note: This inner function runs as thread
        """
        nonlocal packages

        current_python_version = version_parse(
            f"{sys.version_info.major}.{sys.version_info.minor}"
            f".{sys.version_info.micro}"
        )
        consolidated_requirements = SpecifierSet()
        if package_name in requirements:
            for _, specifier in requirements[package_name].items():
                consolidated_requirements &= specifier

        package = packages[package_name]
        current_version = version_parse(package.current)
        current_is_prerelease = (
            str(current_version) == str(package.current)
            and current_version.is_prerelease
        )
        package_name_with_case = package.distribution.metadata["Name"]
        logger.info(
            f"Fetching info for distribution package '{package_name_with_case}' "
            "from PyPI"
        )
        r = requests.get(
            f"https://pypi.org/pypi/{package_name_with_case}/json", timeout=(5, 30)
        )
        if r.status_code == requests.codes.ok:
            pypi_info = r.json()
            latest = ""
            for release, release_details in pypi_info["releases"].items():
                release_detail = (
                    release_details[-1] if len(release_details) > 0 else None
                )
                if not release_detail or (
                    not release_detail["yanked"]
                    and (
                        "requires_python" not in release_detail
                        or not release_detail["requires_python"]
                        or current_python_version
                        in SpecifierSet(release_detail["requires_python"])
                    )
                ):
                    my_release = version_parse(release)
                    if str(my_release) == str(release) and (
                        current_is_prerelease or not my_release.is_prerelease
                    ):
                        if len(consolidated_requirements) > 0:
                            is_valid = my_release in consolidated_requirements
                        else:
                            is_valid = True

                        if is_valid and (
                            not latest or my_release > version_parse(latest)
                        ):
                            latest = release

            if not latest:
                logger.warning(
                    f"Could not find a release of '{package_name_with_case}' "
                    f"that matches all requirements: '{consolidated_requirements}''"
                )
        else:
            if r.status_code == 404:
                logger.info(
                    f"Package '{package_name_with_case}' is not registered in PyPI"
                )
            else:
                logger.warning(
                    "Failed to retrive infos from PyPI for "
                    f"package '{package_name_with_case}'. "
                    f"Status code: {r.status_code}, "
                    f"response: {r.content}"
                )
            latest = ""

        packages[package_name].latest = latest

    if use_threads:
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=MAX_THREAD_WORKERS
        ) as executor:
            executor.map(thread_update_latest_from_pypi, packages.keys())
    else:
        for package_name in packages.keys():
            thread_update_latest_from_pypi(package_name)



from unittest.mock import patch

from django.test import TestCase

from .. import tasks


class TestTasks(TestCase):
    @patch("package_monitor.tasks.Distribution.objects.update_all")
    def test_shoud_update_all_distributions(self, update_all):
        # when
        tasks.update_distributions()
        # then
        self.assertTrue(update_all.called)
